(** * Natural transformations and lax natural transformations *)
Require Import base ord ord.discrete category pcategory functor.
Set Universe Polymorphism.

(** ** Natural transformations between functors *)

Polymorphic Cumulative Structure nat {A B : cat} (F G : functor A B) := mk_nat
  { components :> forall a : A, B (F a) (G a)
  ; _ : forall (a b : A) (f : A a b),
      map F f ∘ components b = components a ∘ map G f
  }.
Arguments mk_nat {A B F G} eta _ : rename.

(** ** Laws *)

Lemma nat_eq (A B : cat) (F G : functor A B) (eta eps : nat F G) :
  components eta = components eps -> eta = eps.
Proof.
  destruct eta,eps => /= E. destruct E. f_equal. exact: pi.
Qed.

Lemma natP (A B : cat) (F G : functor A B) (eta : nat F G) (a b : A) (f : hom a b) :
  map F f ∘ eta b = eta a ∘ map G f.
Proof. by case: eta. Qed.

Lemma natP_comp (A B : cat) (F G : functor A B) (eta : nat F G)
      (a b : A) (c : B) (f : hom a b) (g : hom (G b) c) :
  map F f ∘ (eta b ∘ g) = eta a ∘ (map G f ∘ g).
Proof. by rewrite <- compA, natP, compA. Qed.

Hint Rewrite natP natP_comp : cat.

(** ** Functor Category *)

Section FunctorCategory.
  Variables (A B : cat).
  Implicit Types (F G H : functor A B).

  Program Definition identity_nat F : nat F F := mk_nat (fun _ => cid) _.
  Next Obligation. by csimpl. Qed.

  Program Definition compose_nat F G H (eta : nat F G) (eps : nat G H) : nat F H :=
    mk_nat (fun a => eta a ∘ eps a) _.
  Next Obligation. by csimpl. Qed.

  Lemma nat_idL F G (eta : nat F G) :
    compose_nat (identity_nat F) eta = eta.
  Proof.
    apply: nat_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma nat_idR F G (eta : nat F G) :
    compose_nat eta (identity_nat G) = eta.
  Proof.
    apply: nat_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma nat_compA F G H I (e1 : nat F G) (e2 : nat G H) (e3 : nat H I) :
    compose_nat (compose_nat e1 e2) e3 = compose_nat e1 (compose_nat e2 e3).
  Proof.
    apply: nat_eq; fext=>/=a. by csimpl.
  Qed.

  Canonical functor_catMixin := CatMixin nat_idL nat_idR nat_compA.
  Canonical FUNCTOR := Eval hnf in Cat (functor A B) nat functor_catMixin.
End FunctorCategory.

(** ** Natural transformations between (2,1)-functors *)

Definition pnat {A B : pcat} (F G : pfunctor A B) := nat F G.

(** ** Order on natural transformations *)

Section NatOrder.
  Variables (A B : pcat) (F G : pfunctor A B).
  Implicit Types (eta eps psi : pnat F G).

  Definition nat_le eta eps := forall a, eta a ≤ eps a.

  Lemma nat_le_refl eta : nat_le eta eta. by []. Qed.
  Lemma nat_le_trans : transitive nat_le.
  Proof. move=> eta eps psi l1 l2 a. rewrite (l1 a). exact: l2. Qed.

  Canonical nat_le_proMixin := ProMixin nat_le_refl nat_le_trans.
  Canonical nat_proType := Eval hnf in ProType (nat F G) nat_le_proMixin.

  Lemma nat_le_eq : antisymmetric nat_le.
  Proof.
    move=> eta eps l1 l2. apply: nat_eq. fext=> a.
    apply: le_eq. exact: l1. exact: l2.
  Qed.

  Canonical nat_le_ordMixin := OrdMixin nat_le_eq.
  Canonical nat_ordType := Eval hnf in OrdType (nat F G) nat_le_ordMixin.
End NatOrder.

(** ** (2,1)-functor Category *)

Section PFunctorCategory.
  Variables (A B : pcat).
  Implicit Types (F G H : pfunctor A B).

  Canonical pfunctor_catMixin := @CatMixin (pfunctor A B) pnat
    (@identity_nat A B) (@compose_nat A B) (@nat_idL A B) (@nat_idR A B) (@nat_compA A B).
  Canonical pfunctor_cat := Eval hnf in Cat (pfunctor A B) pnat pfunctor_catMixin.

  Global Instance compose_nat_proper F G H :
    Proper (@ord_op (nat_ordType F G) ++>
            @ord_op (nat_ordType G H) ++>
            @ord_op (nat_ordType F H)) (@compose_nat A B F G H).
  Proof. move=> e1 e2 l1 f1 f2 l2 a /=. by rewrite (l1 a) (l2 a). Qed.

  Canonical PFUNCTOR := @Ord2.Pack (pfunctor A B) pnat
                                   (@Ord2.Class (pfunctor A B) pnat
                                                pfunctor_catMixin
                                                (fun F G => _)
                                                compose_nat_proper)
                                   (pfunctor A B) pnat.
  (* universe anomaly! *)
  (* Canonical PFUNCTOR := Eval hnf in PCat (pfunctor A B) pnat. *)
End PFunctorCategory.

(** ** Lax-natural transformations between (2,1)-functors *)

Polymorphic Cumulative Structure lax {A B : pcat} (F G : pfunctor A B) := mk_lax
  { lax_components :> forall a : @Ord2.sort A, @Ord2.hom B (F a) (G a)
  ; _ : forall (a b : A) (f : hom a b),
      (map F f ∘ lax_components b) ≤ (lax_components a ∘ map G f) :> hom (F a) (G b)
  }.
Arguments mk_lax {A B F G} eta _ : rename.

(** ** Laws *)

Lemma lax_eq (A B : pcat) (F G : pfunctor A B) (eta eps : lax F G) :
  lax_components eta = lax_components eps -> eta = eps.
Proof.
  destruct eta,eps => /= E. destruct E. f_equal. exact: pi.
Qed.

Lemma laxP (A B : pcat) (F G : pfunctor A B) (eta : lax F G) (a b : A) (f : hom a b) :
  map F f ∘ eta b ≤ eta a ∘ map G f.
Proof. by case: eta. Qed.

Lemma laxP_comp (A B : pcat) (F G : pfunctor A B) (eta : lax F G)
      (a b : A) (c : B) (f : hom a b) (g : hom (G b) c) :
  map F f ∘ (eta b ∘ g) ≤ eta a ∘ (map G f ∘ g).
Proof. by rewrite <- compA, laxP, compA. Qed.

Hint Rewrite laxP laxP_comp : cat.

(** ** Order on lax natural transformations *)

Section LaxOrder.
  Variables (A B : pcat) (F G : pfunctor A B).
  Implicit Types (eta eps psi : lax F G).

  Definition lax_le eta eps := forall a, eta a ≤ eps a.

  Lemma lax_le_refl eta : lax_le eta eta. by []. Qed.
  Lemma lax_le_trans : transitive lax_le.
  Proof. move=> eta eps psi l1 l2 a. rewrite (l1 a). exact: l2. Qed.

  Canonical lax_le_proMixin := ProMixin lax_le_refl lax_le_trans.
  Canonical lax_proType := Eval hnf in ProType (lax F G) lax_le_proMixin.

  Lemma lax_le_eq : antisymmetric lax_le.
  Proof.
    move=> eta eps l1 l2. apply: lax_eq. fext=> a.
    apply: le_eq. exact: l1. exact: l2.
  Qed.

  Canonical lax_le_ordMixin := OrdMixin lax_le_eq.
  Canonical lax_ordType := Eval hnf in OrdType (lax F G) lax_le_ordMixin.
End LaxOrder.

(** ** Lax (2,1)-Functor (2,1)-Category *)

Section LaxFunctorPCategory.
  Variables (A B : pcat).
  Implicit Types (F G H : pfunctor A B).

  Program Definition identity_lax F : lax F F := mk_lax (fun _ => cid) _.
  Next Obligation. by csimpl. Qed.

  Program Definition compose_lax F G H (eta : lax F G) (eps : lax G H) : lax F H :=
    mk_lax (fun a => eta a ∘ eps a) _.
  Next Obligation. by csimpl. Qed.

  Lemma lax_idL F G (eta : lax F G) :
    compose_lax (identity_lax F) eta = eta.
  Proof.
    apply: lax_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma lax_idR F G (eta : lax F G) :
    compose_lax eta (identity_lax G) = eta.
  Proof.
    apply: lax_eq; fext=>/=a. by csimpl.
  Qed.

  Lemma lax_compA F G H I (e1 : lax F G) (e2 : lax G H) (e3 : lax H I) :
    compose_lax (compose_lax e1 e2) e3 = compose_lax e1 (compose_lax e2 e3).
  Proof.
    apply: lax_eq; fext=>/=a. by csimpl.
  Qed.

  Canonical lax_functor_catMixin := CatMixin lax_idL lax_idR lax_compA.
  Canonical lax_functor_cat := Eval hnf in Cat (pfunctor A B) lax lax_functor_catMixin.

  Global Instance compose_lax_proper F G H :
    Proper (ord_op ++> ord_op ++> ord_op) (@compose_lax F G H).
  Proof. move=> /=e1 e2 l1 f1 f2 l2 a /=. by rewrite (l1 a) (l2 a). Qed.

  (* Inference problems *)
  Canonical LFUNCTOR := @Ord2.Pack (pfunctor A B) lax
                                   (@Ord2.Class (pfunctor A B) lax
                                                lax_functor_catMixin
                                                (fun F G => _)
                                                compose_lax_proper)
                                   (pfunctor A B) lax.
  (*Canonical LFUNCTOR := Eval hnf in PCat (pfunctor A B) lax.*)
End LaxFunctorPCategory.

(** ** Functor to Lax Functor *)

Section ToLax.
  Variables (A B : pcat).

  Program Definition to_lax (F G : pfunctor A B) (eps : pnat F G) : lax F G :=
    mk_lax eps _.
  Next Obligation. rewrite-> natP. reflexivity. Qed.

  Program Definition to_lax_functor : functor (PFUNCTOR A B) (LFUNCTOR A B) :=
    mk_functor id (@to_lax) _ _.
  Next Obligation. exact: lax_eq. Qed.
  Next Obligation. exact: lax_eq. Qed.

  Global Instance to_lax_proper F G :
    Proper (@ord_op (nat_ordType F G) ++> ord_op) (@to_lax F G).
  Proof. move=> eps eta le_eps_eta a /=. exact: le_eps_eta. Qed.

  Definition TO_LAX : pfunctor (PFUNCTOR A B) (LFUNCTOR A B) :=
    mk_pfunctor to_lax_functor.
End ToLax.
