(** * (2,1)-Categories *)
Require Import base ord ord.discrete category.
Set Universe Polymorphism.

Module Ord2.
  Polymorphic Definition mixin_of (T : Type) (hom : T -> T -> ordType)
    (comp : forall a b c, hom a b -> hom b c -> hom a c) := forall a b c,
    Proper (@ord_op (hom a b) ++> @ord_op (hom b c) ++> @ord_op (hom a c)) (@comp a b c).

  Section ClassDef.
    Polymorphic Cumulative Record class_of (T : Type) (hom : T -> T -> Type) := Class {
      cat_base : @Cat.class_of T hom;
      hom_base : forall a b, Ord.class_of (hom a b);
      mixin :
        @mixin_of T
          (fun a b => @Ord.Pack (hom a b) (hom_base a b) (hom a b))
          (@Cat.comp T hom cat_base)
    }.
    Local Coercion cat_base : class_of >-> Cat.class_of.

    Structure type := Pack
      {sort; hom; _ : @class_of sort hom; sort' : Type; _ : sort' -> sort' -> Type }.
    Local Coercion sort : type >-> Sortclass.
    Local Coercion hom : type >-> Funclass.

    Variable (T : Type) (hom : T -> T -> Type) (cT : type).
    Definition class := let: Pack _ _ c _ _ as cT' := cT return class_of cT' in c.
    Definition clone c of phant_id class c := @Pack T hom c T hom.
    Let xT := let: Pack T _ _ _ _ := cT in T.
    Let xF := let: Pack _ hom _ _ _ := cT return cT -> cT -> Type in hom.
    Notation xclass := (class : @class_of xT xF).

    Definition pack :=
      (* Find the underlying category *)
      [find ct | @Cat.hom ct ~ hom | "is not the type of morphisms of a category"]
      fun (_ : unify (@Cat.sort ct) T (Some "is not the type of objects of a category")) =>
      (*[find ct | Cat.sort ct ~ T | "is not the type of objects of a category"]
      fun (_ : unify (@Cat.hom ct) hom (Some "is not the type of morphisms of a category")) =>*)
      (* Find the canonical category mixin *)
      [find ce | Cat.class ct ~ ce]
      (* Find the order on homsets *)
      [find ot | (fun (a b : T) => Ord.sort (ot a b)) ~ hom | "is not an ordType"]
      [find oe | (fun a b => Ord.class (ot a b)) ~ (oe : forall a b, Ord.class_of (hom a b))]
      (* Use typeclass inference to show monotonicity *)
      fun (mixin : forall a b c,
              Proper
                (@ord_op (@Pro.Pack (hom a b) (oe a b) (hom a b)) ++>
                 @ord_op (@Pro.Pack (hom b c) (oe b c) (hom b c)) ++>
                 @ord_op (@Pro.Pack (hom a c) (oe a c) (hom a c)))
                (@Cat.comp T hom ce a b c)) =>
        @Pack T hom (@Class T hom ce oe mixin) T hom.

    Definition cat := @Cat.Pack cT cT xclass xT xF.

    Section OrdType.
      Variable (a b : cT).
      Definition proType := @Pro.Pack (cT a b) (hom_base class a b) (xF a b).
      Definition ordType := @Ord.Pack (cT a b) (hom_base class a b) (xF a b).
    End OrdType.
  End ClassDef.

  Module Exports.
    Coercion cat_base : class_of >-> Cat.class_of.
    Coercion mixin : class_of >-> mixin_of.
    Coercion sort : type >-> Sortclass.
    Coercion hom : type >-> Funclass.

    Coercion cat : type >-> Cat.type.
    Canonical cat.
    Canonical proType.
    Canonical ordType.

    Notation pcat := type.
    Notation PCat T hom := (@pack T hom _ idfun idfun _ idfun _ idfun _ idfun _).

    Notation "[ 'pcat' 'of' hom 'for' C ]" := (@clone _ hom C _ idfun)
      (at level 0, format "[ 'pcat'  'of'  hom  'for'  C ]") : form_scope.
    Notation "[ 'pcat' 'of' hom ]" := (@clone _ hom _ _ id)
      (at level 0, format "[ 'pcat'  'of'  hom ]") : form_scope.
  End Exports.
End Ord2.
Export Ord2.Exports.

(** ** Laws *)

Global Instance comp_proper (C : pcat) (a b c : C) :
  Proper (ord_op ++> ord_op ++> ord_op) (@comp C a b c) :=
  @Ord2.mixin _ _ (Ord2.class C) a b c.

(** ** Opposite of a (2,1)-category *)

Section OP.
  Variable C : pcat.

  Global Instance op_comp_proper (a b c : C^op) :
    Proper (ord_op ++> ord_op ++> ord_op) (@comp (OP C) a b c).
  Proof. move=> a1 a2 l1 b1 b2 l2. csimpl. by rewrite l1 l2. Qed.

  Canonical POP := Eval hnf in PCat (C^op) (@op_hom C).
End OP.

(** ** Category as a discrete (2,1)-category *)

Section DISC.
  Variable C : cat.
  Definition discrete_obj : Type := C.
  Definition discrete_hom (a b : C) : Type := Δ (C a b).

  Canonical discrete_catMixin := @CatMixin discrete_obj discrete_hom
    (@cid C) (@comp C) (@idL C) (@idR C) (@compA C).
  Canonical discrete_cat :=
    Eval hnf in Cat discrete_obj discrete_hom discrete_catMixin.

  (* Setoid is clever enough that we don't need this instance. *)
  (*Global Instance discrete_comp_proper (a b c : discrete_cat) :
    Proper (ord_op ++> ord_op ++> ord_op) (@comp discrete_cat a b c).
  Proof. move=> a1 a2 l1 b1 b2 l2. destruct l1, l2. reflexivity. Qed.*)

  Canonical DISC := Eval hnf in PCat discrete_obj discrete_hom.
End DISC.

(** ** (2,1)-Category of Posets *)

Section POSET.
  Implicit Types (A B C D : ordType).

  Definition ord_mfun_type (A B : ordType) : Type :=
    {mono A -> B}.

  Definition mfun_id A : {mono A -> A} :=
    @mfun _ _ idfun (fun _ _ => idfun).

  Program Definition mfun_comp A B C (f : {mono A -> B}) (g : {mono B -> C}) : {mono A -> C} :=
    mk_mfun (g \o f) _.
  Next Obligation. by rewrite H. Qed.

  Lemma mfun_idL A B (f : {mono A -> B}) :
    mfun_comp (mfun_id A) f = f.
  Proof. exact: mfun_eq. Qed.

  Lemma mfun_idR A B (f : {mono A -> B}) :
    mfun_comp f (mfun_id B) = f.
  Proof. exact: mfun_eq. Qed.

  Lemma mfun_compA A B C D (f : {mono A -> B}) (g : {mono B -> C}) (h : {mono C -> D}) :
    mfun_comp (mfun_comp f g) h = mfun_comp f (mfun_comp g h).
  Proof. exact: mfun_eq. Qed.

  Global Instance mfun_comp_proper (A B C : ordType) :
    Proper (ord_op ++> ord_op ++> ord_op) (@mfun_comp A B C).
  Proof.
    move=> f1 f2 l1 g1 g2 l2 x /=. by rewrite (l2 _) (l1 _).
  Qed.

  Canonical ordType_catMixin := @CatMixin ordType ord_mfun_type
    (@mfun_id) (@mfun_comp) mfun_idL mfun_idR mfun_compA.
  Canonical ordType_cat :=
    Eval hnf in Cat ordType ord_mfun_type ordType_catMixin.
  Canonical POSET :=
    Eval hnf in PCat ordType ord_mfun_type.
End POSET.
