(** ** The 2-poset of lax presheafs on the category of renamings *)
Require Import base ord data.fintype.

(*Set Primitive Projections.*)

Polymorphic Cumulative Structure category := mk_category
  { obj :> Type
  ; hom :> obj -> obj -> Type
  ; idc : forall a, hom a a
  ; comp : forall a b c, hom a b -> hom b c -> hom a c
  ; idL : forall a b (f : hom a b), comp (idc a) f = f
  ; idR : forall a b (f : hom a b), comp f (idc b) = f
  ; compA : forall a b c d (f : hom a b) (g : hom b c) (h : hom c d),
      comp (comp f g) h = comp f (comp g h)
  ; hom_ordClass : forall a b, Ord.class_of (hom a b)
  ; comp_proper' : forall a b c,
      Proper (@ord_op (Pro.Pack (hom_ordClass a b) (hom a b)) ++>
              @ord_op (Pro.Pack (hom_ordClass b c) (hom b c)) ++>
              @ord_op (Pro.Pack (hom_ordClass a c) (hom a c))) (@comp a b c)
  }.
Arguments idc {C a} : rename.
Arguments comp {C a b c} f g : rename.
Arguments hom {C} a b : rename.
Notation "f ∘ g" := (comp f g) : cat_scope.
Open Scope cat_scope.

(** ** Canonical order on homsets *)
Section HomOrdType.
  Variable (A : category) (a b : A).
  Canonical hom_proType := @Pro.Pack (hom a b) (@hom_ordClass A a b) (hom a b).
  Canonical hom_ordType := @Ord.Pack (hom a b) (@hom_ordClass A a b) (hom a b).
End HomOrdType.

Global Instance comp_proper (A : category) (a b c : A) :
  Proper (ord_op ++> ord_op ++> ord_op) (@comp A a b c).
Proof. exact: comp_proper'. Qed.

Polymorphic Cumulative Structure domain (A : category) := mk_domain
  { domain_fam :> A -> ordType
  ; th : forall (a b : A) (f : A a b), domain_fam a -> domain_fam b
  ; th1 : forall a (x : domain_fam a),
      th idc x = x
  ; thC : forall a b c (f : A a b) (g : A b c) x,
      th g (th f x) = th (f ∘ g) x
  ; th_proper : forall a b, Proper (ord_op ++> ord_op ++> ord_op) (@th a b)
  }.

Polymorphic Cumulative Structure lax {A : category} (F G : domain A) := mk_lax
  { lax_components :> forall a, F a -> G a
  ; lax_mono : forall a (x y : F a), x ≤ y -> lax_components x ≤ lax_components y
  ; laxP : forall a b (f : A a b) (x : F a),
      lax_components (th f x) ≤ th f (lax_components x)
  }.

Global Instance lax_proper {A : category} {F G : domain A} (eps : lax F G) (a : A) :
  Proper (ord_op ++> ord_op) (@lax_components A F G eps a).
Proof. move=> x y le_x_y. exact: lax_mono. Qed.

(*Global Instance lax_mono' {A : category} {F G : domain A} (eps : lax F G) (a : A) :
  monotone (@lax_components A F G eps a).
Proof. exact _. Qed.*)

(*Definition R : category :=
  {| obj := nat
   ; hom := ren
   ; idc := fun _ => id
   ; comp := fun _ _ _ f g => f >> g
   ; idL := fun _ _ _ => erefl
   ; idR := fun _ _ _ => erefl
   ; compA := fun _ _ _ _ _ _ _ => erefl
  |}.

(** ** Discrete ordTypes *)

Section DiscreteOrder.
  Variable A : Type.
  Definition discrete := A.
  Implicit Types (x y z : discrete).

  Definition discrete_le x y := (x = y).

  Lemma discrete_le_refl : reflexive discrete_le.
  Proof. by []. Qed.

  Lemma discrete_le_trans : transitive discrete_le.
  Proof. by move=> x y z ->. Qed.

  Canonical discrete_proMixin := ProMixin discrete_le_refl discrete_le_trans.
  Canonical discrete_proType := Eval hnf in ProType discrete discrete_proMixin.

  Lemma discrete_le_eq : antisymmetric discrete_le.
  Proof. by move=> x y ->. Qed.

  Canonical discrete_ordMixin := OrdMixin discrete_le_eq.
  Canonical discrete_ordType := Eval hnf in OrdType discrete discrete_ordMixin.
End DiscreteOrder.

Notation "'Δ' A" := (discrete A) (at level 2).

(** ** Codiscrete proType *)

Section CodiscreteProType.
  Variable A : Type.
  Definition codiscrete := A.
  Implicit Types (x y z : codiscrete).

  Definition codiscrete_le x y := True.

  Lemma codiscrete_refl : reflexive codiscrete_le.
  Proof. by []. Qed.

  Lemma codiscrete_trans : transitive codiscrete_le.
  Proof. by []. Qed.

  Canonical codiscrete_proMixin := ProMixin codiscrete_refl codiscrete_trans.
  Canonical codiscrete_proType := Eval hnf in ProType codiscrete codiscrete_proMixin.
End CodiscreteProType.

Notation "'∇' A" := (codiscrete A) (at level 2).


Program Definition Fin : domain R :=
  {| domain_fam := fun n => [ordType of Δ (fin n)]
   ; th := fun _ _ => id
   ; th1 := fun _ _ => erefl
   ; thC := fun _ _ _ _ _ _ => erefl
  |}.
Next Obligation. by destruct H. Qed.

(** ** Product order *)

Section ProductProType.
  Variable (A B : proType).
  Implicit Types (p q r : A * B).

  Definition prod_le p q := (p.1 ≤ q.1) /\ (p.2 ≤ q.2).

  Lemma prod_le_refl : reflexive prod_le.
  Proof. by []. Qed.

  Lemma prod_le_trans : transitive prod_le.
  Proof.
    move=> p q r [l11 l12] [l21 l22]. split; by [rewrite l11|rewrite l12].
  Qed.

  Canonical prod_proMixin := ProMixin prod_le_refl prod_le_trans.
  Canonical prod_proType := Eval hnf in ProType (A * B) prod_proMixin.
End ProductProType.

Section ProductOrdType.
  Variable (A B : ordType).
  Implicit Types (p q : A * B).

  Lemma prod_le_eq p q : p ≤ q -> q ≤ p -> p = q.
  Proof.
    move: p q => -[p1 p2] [q1 q2] [/=h11 h12] [/=h21 h22].
    congr pair; exact: le_eq.
  Qed.

  Canonical prod_ordMixin := OrdMixin prod_le_eq.
  Canonical prod_ordType := Eval hnf in OrdType (A * B) prod_ordMixin.
End ProductOrdType.

(* TODO: clat, frame structure is not needed, just nice to have *)
(*Section ProductCLat.
  Variable (A B : clat).
  Implicit Types (p q : A * B).

  Definition *)

*)
(** ** Lax ord type *)

Section LaxOrdType.
  Context {A : category} (F G : domain A).
  Implicit Types (eta eps xi : lax F G).

  Definition lax_le eta eps := forall a, eta a ≤ eps a.

  Lemma lax_le_refl : reflexive lax_le.
  Proof. move=> eps a. by []. Qed.

  Lemma lax_le_trans : transitive lax_le.
  Proof.
    move=> eps eta xi le_eps_eta le_eta_xi a.
      by rewrite (le_eps_eta a) (le_eta_xi a).
  Qed.

  Canonical lax_proMixin := ProMixin lax_le_refl lax_le_trans.
  Canonical lax_proType := Eval hnf in ProType (lax F G) lax_proMixin.

  Lemma lax_eq eta eps : lax_components eta = lax_components eps -> eta = eps.
  Proof. destruct eta,eps => /= E. destruct E. f_equal; exact: pi. Qed.

  Lemma lax_le_eq eta eps : eta ≤ eps -> eps ≤ eta -> eta = eps.
  Proof.
    move=> l1 l2. apply: lax_eq. apply: fext => a.
    apply: le_eq. exact: l1. exact: l2.
  Qed.

  Canonical lax_ordMixin := OrdMixin lax_le_eq.
  Canonical lax_ordType := Eval hnf in OrdType (lax F G) lax_ordMixin.
End LaxOrdType.

(** ** The category of domains *)

Section DomainCategory.
  Variable (A : category).

  Definition lax_id (F : domain A) : lax F F :=
    {| lax_components := fun x => id
     ; lax_mono := fun _ _ _ => id
     ; laxP := fun _ _ _ _ => le_refl
    |}.

  Program Definition lax_comp {F G H : domain A} (eta : lax F G) (eps : lax G H) : lax F H :=
    {| lax_components := fun a => eta a >> eps a |}.
  Next Obligation. apply: mono. exact: mono. Qed.
  Next Obligation. by rewrite !laxP. Qed.

  Canonical Domain : category.
  refine ({| obj := domain A; hom := lax; idc := lax_id; comp := @lax_comp |}).
    abstract (intros; exact: lax_eq).
    abstract (intros; exact: lax_eq).
    abstract (intros; exact: lax_eq).
    Unshelve. Focus 2. intros a b. exact (Ord.class [ordType of (lax a b)]).
    abstract (move=> F G H /= eta1 eta2 le1 eps1 eps2 le2 a x /=; by rewrite (le1 a x) (le2 a _)).
  Defined.
End DomainCategory.

(** ** Products in the category of domains *)

Section ProductDomain.
  Context {A : category}.
  Implicit Types (F G : domain A).

  Program Definition Prod F G : domain A :=
    {| domain_fam := fun a => [ordType of (F a * G a)%type]
     ; th := fun a b f p => (th f p.1, th f p.2)
    |}.
  Next Obligation. move=>/=. congr pair; exact: th1. Qed.
  Next Obligation. move=>/=. congr pair; exact: thC. Qed.
  Next Obligation. move=>/= f g le1 x y [le2 le3]. split; exact: th_proper. Qed.

  Program Definition lax_pi1 {F G} : lax (Prod F G) F :=
    {| lax_components := fun a p => p.1 |}.
  Next Obligation. by case: H. Qed.

  Program Definition lax_pi2 {F G} : lax (Prod F G) G :=
    {| lax_components := fun a p => p.2 |}.
  Next Obligation. by case: H. Qed.

  Program Definition lax_pair {F G H} (eps : lax F G) (eta : lax F H) : lax F (Prod G H) :=
    {| lax_components := fun a x => (eps a x, eta a x) |}.
  Next Obligation. split=>/=; by rewrite H0. Qed.
  Next Obligation. split=>/=; exact: laxP. Qed.

  Lemma lax_pair1 F G H (eps : lax F G) (eta : lax F H) :
    lax_pair eps eta ∘ lax_pi1 = eps.
  Proof. exact: lax_eq. Qed.

  Lemma lax_pair2 F G H (eps : lax F G) (eta : lax F H) :
    lax_pair eps eta ∘ lax_pi2 = eta.
  Proof. exact: lax_eq. Qed.

  Lemma lax_pair_eta F G H (eta : lax F (Prod G H)) :
    eta = lax_pair (eta ∘ lax_pi1) (eta ∘ lax_pi2).
  Proof.
    apply: lax_eq. fext=> a x /=. by case: (eta a x).
  Qed.

  Definition lax_first {F G H} (eta : lax F G) : lax (Prod F H) (Prod G H) :=
    lax_pair (lax_pi1 ∘ eta) lax_pi2.
End ProductDomain.

Section Yoneda.
  Context {A : category} (a : A).

  Program Definition Yoneda : domain A :=
    {| domain_fam := fun b => [ordType of hom a b] 
     ; th := fun a b f g => g ∘ f
    |}.
  Next Obligation. exact: idR. Qed.
  Next Obligation. exact: compA. Qed.
  Next Obligation. move=> f1 f2 le1 g1 g2 le2. by rewrite le1 le2. Qed.

  Definition yoneda {F : domain A} (eta : hom Yoneda F) : F a := eta a idc.

  Program Definition yonedaI {F : domain A} (x : F a) : hom Yoneda F :=
    {| lax_components := fun _ f => th f x |}.
  Next Obligation. exact: th_proper. Qed.
  Next Obligation. by rewrite thC. Qed.

  Lemma yoneda_sect {F : domain A} (eta : hom Yoneda F) :
    eta ≤ yonedaI (yoneda eta).
  Proof.
    move=> /=x f. rewrite/yoneda. by rewrite -laxP /= idL.
  Qed.

  Lemma yoneda_retr {F : domain A} (x : F a) :
    yoneda (yonedaI x) = x.
  Proof. move=>/=. exact: th1. Qed.
End Yoneda.

Section YonedaMap.
  Context {A : category}.

  Program Definition ymap {a b : A} (f : hom a b) : hom (Yoneda b) (Yoneda a) :=
    {| lax_components := fun c g => f ∘ g |}.
  Next Obligation. by rewrite H. Qed.
  Next Obligation. by rewrite compA. Qed.

  Lemma ymap1 a : ymap (idc : hom a a) = idc.
  Proof. apply: lax_eq. fext=> b /= f. exact: idL. Qed.

  Lemma ymapC a b c (f : hom a b) (g : hom b c) :
    ymap g ∘ ymap f = ymap (f ∘ g).
  Proof. apply: lax_eq. fext=> d /= h. by rewrite compA. Qed.
End YonedaMap.

Section LaxExponentialDomain.
  Context {A : category} (F G : domain A).

  Program Definition Exp : domain A :=
    {| domain_fam := fun a => [ordType of lax (Prod (Yoneda a) F) G]
     ; th := fun a b (f : hom a b) (g : lax (Prod (Yoneda a) F) G) =>
               lax_first (ymap f) ∘ g
    |}.
  Next Obligation.
    apply: lax_eq; fext=> b [/=f y]. by rewrite idL.
  Qed.
  Next Obligation.
    apply: lax_eq; fext=> d [/=h y]. by rewrite compA.
  Qed.
  Next Obligation.
    move=> f1 f2 le1 /= g1 g2 le2 c [/=h x]. rewrite (le2 _ _).
    apply: lax_mono. split=> //=. by rewrite le1.
  Qed.
End LaxExponentialDomain.

