Require Import base data.fintype.
Require Import pts.category.
Set Universe Polymorphism.
Set Printing Universes.

(** * The category of contexts *)

Notation ℕ := nat.
Definition ctx : Type := ℕ.
Definition cmorph (m n : ctx) : Type := fin n -> fin m.
Definition ctxid (m : ctx) : cmorph m m := id.
Definition ccomp m n k (f : cmorph m n) (g : cmorph n k) : cmorph m k := g >> f.

Lemma ccomp_idL m n (f : cmorph m n) : ccomp (@ctxid _) f = f.
Proof. by []. Qed.

Lemma ccomp_idR m n (f : cmorph m n) : ccomp f (@ctxid _) = f.
Proof. by []. Qed.

Lemma ccomp_assoc m n k l (f : cmorph m n) (g : cmorph n k) (h : cmorph k l) :
  ccomp (ccomp f g) h = ccomp f (ccomp g h).
Proof. by []. Qed.

Canonical ctx_catMixin := @CatMixin ctx cmorph ctxid ccomp
  ccomp_idL ccomp_idR ccomp_assoc.
Canonical CTX := Eval hnf in Cat ctx cmorph ctx_catMixin.

(** *** Products in contexts *)

Definition ctx_prod (m n : ctx) : ctx := m + n.
Definition ctx_terminal : ctx := 0.

Fixpoint fin_split m : forall n, fin (m + n) -> fin m + fin n :=
  match m as m return forall n, fin (m+n) -> (fin m + fin n) with
  | 0 => fun n i => inr i
  | m.+1 => fun n i =>
    match i with
    | None => inl bound
    | Some j =>
      match @fin_split m n j with
      | inl k => inl (shift k)
      | inr k => inr k
      end
    end
  end.

Definition fin_inl m n (i : fin m) : fin (m + n).
Proof.
  induction m; cbn in *. destruct i. destruct i as [j|].
  apply shift. exact (IHm j). exact bound.
Defined.

Definition fin_inr m n (i : fin n) : fin (m + n).
Proof.
  induction m; cbn in *. exact i. apply shift. exact IHm.
Defined.

Lemma fin_inl_split m n (i : fin m) :
  fin_split (@fin_inl m n i) = inl i.
Proof.
  induction m; cbn in *. destruct i. destruct i as [j|]; cbn=>//.
    by rewrite (IHm j).
Qed.

Lemma fin_inr_split m n (i : fin n) :
  fin_split (@fin_inr m n i) = inr i.
Proof.
  induction m; cbn in * => //. by rewrite IHm.
Qed.

Lemma fin_split_eta m n (i : fin (m + n)) :
  match fin_split i with
  | inl j => fin_inl n j
  | inr j => fin_inr m j
  end = i.
Proof.
  induction m; cbn in *=>//. destruct i as [j|]; cbn=>//.
  move: (IHm j). case: (fin_split j) => k eqn.
  cbn. by rewrite -eqn. by rewrite -eqn.
Qed.

Definition ctx_pi1 (m n : ctx) : hom (ctx_prod m n) m.
  move=> i. cbn. exact: fin_inl.
Defined.

Definition ctx_pi2 (m n : ctx) : hom (ctx_prod m n) n.
  move=> i. cbn. exact: fin_inr.
Defined.

Definition ctx_pair (m n k : ctx) (f : hom m n) (g : hom m k) : hom m (ctx_prod n k).
  move=> i. cbn. destruct (fin_split i) as [j|j]. exact (f j). exact (g j).
Defined.

Definition ctx_pair_pi1 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi1 _ _ = f.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. by rewrite fin_inl_split.
Qed.

Definition ctx_pair_pi2 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi2 _ _ = g.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. by rewrite fin_inr_split.
Qed.

Definition ctx_pair_eta (m n k : ctx) (f : hom m (ctx_prod n k)) :
  ctx_pair (f ∘ @ctx_pi1 _ _) (f ∘ @ctx_pi2 _ _) = f.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. rewrite -{2}[i]fin_split_eta.
    by case: (fin_split i).
Qed.

Definition ctx_bang (n : ctx) : hom n ctx_terminal := null.

Lemma ctx_terminal_eta (n : ctx) (f : hom n ctx_terminal) :
  f = ctx_bang n.
Proof. by fext. Qed.

Canonical ctx_cartcatMixin :=
  @CartCat.Mixin ctx cmorph ctx_catMixin
    ctx_terminal ctx_bang ctx_prod ctx_pi1 ctx_pi2 ctx_pair
    ctx_pair_pi1 ctx_pair_pi2 ctx_pair_eta
    ctx_terminal_eta.
Canonical ctx_cartcat :=
  Eval hnf in CartCat.pack ctx_cartcatMixin.

(** * Functors *)

Polymorphic Cumulative Structure Functor (C D : cat) : Type := mk_functor {
  fobj :> C -> D;
  fmap : forall (a b : C), C a b -> D (fobj a) (fobj b);
  fmap_id : forall (a : C), fmap (@cid C a) = cid;
  fmap_comp : forall (a b c : C) (f : C a b) (g : C b c),
      fmap f ∘ fmap g = fmap (f ∘ g)
}.

(** * Category of Elements *)

Section CategoryOfElements.
  Variable (A : cat) (P : presheaf A).

  Definition el_obj := { a : A & P a }.
  Definition el_hom (a b : el_obj) :=
    { f : A a.1 b.1 | map f b.2 = a.2 }.

  Definition el_hom_id (a : el_obj) : el_hom a a.
    refine (exist _ cid _). exact: map1.
  Defined.

  Definition el_hom_comp (a b c : el_obj) (f : el_hom a b) (g : el_hom b c) :
    el_hom a c.
    refine (exist _ (f.1 ∘ g.1) _). abstract by rewrite <- mapC, g.2, f.2.
  Defined.

  Lemma el_hom_comp_idL (a b : el_obj) (f : el_hom a b) :
    el_hom_comp (el_hom_id a) f = f.
  Proof. apply: sig_eq => /=. by csimpl. Qed.

  Lemma el_hom_comp_idR (a b : el_obj) (f : el_hom a b) :
    el_hom_comp f (el_hom_id b) = f.
  Proof. apply: sig_eq => /=. by csimpl. Qed.

  Lemma el_hom_comp_assoc (a b c d : el_obj) (f : el_hom a b) (g : el_hom b c) (h : el_hom c d) :
    el_hom_comp (el_hom_comp f g) h = el_hom_comp f (el_hom_comp g h).
  Proof. apply: sig_eq => /=. by csimpl. Qed.

  Canonical el_catMixin := @CatMixin el_obj el_hom el_hom_id el_hom_comp
    el_hom_comp_idL el_hom_comp_idR el_hom_comp_assoc.
  Canonical EL := Eval hnf in Cat el_obj el_hom el_catMixin.
End CategoryOfElements.

Definition emap {A : cat} {P : presheaf A} {a b : A} (f : A a b) (x : P b) :
  el_hom (a; map f x) (b; x).
Proof. exists f. cbn. exact erefl. Defined.

Unset Printing Universes.

Lemma emap1 {A : cat} {P : presheaf A} {a : A} (x : P a) :
  @cid (EL P) _ = transport (map1 x) (P := fun y => el_hom (a; y) (a; x)) (emap cid x).
Proof. apply: sig_eq. cbn. by destruct (map1 x). Qed.

Lemma emapC {A : cat} {P : presheaf A} {a b c : A} (f : A a b) (g : A b c) (x : P c) :
  emap (f ∘ g) x = transport (mapC f g x) (P := fun y => el_hom (_; y) _)
                             (emap f (map g x) ∘ emap g x).
Proof.
  apply: sig_eq. cbn. by destruct (mapC f g x).
Qed.

(** * Presheaf Containers *)

Polymorphic Cumulative Structure Container (C D : cat) := mk_container {
  shape : presheaf D;
  positions : Functor (EL shape) (PRESHEAF C)
}.

Obligation Tactic := idtac.

(* Some interesting dependent types in here: *)
Program Definition ext {C D : cat} (SP : Container C D) (P : presheaf C) : presheaf D :=
  mk_presheaf
    (fun d : D => { s : shape SP d & pnat (positions SP (d; s)) P })
    (fun a b f el => (map f el.1; fmap (positions SP) (emap f el.1) ∘ el.2))
    _ _.
Next Obligation.
  move=> C D SP P a [s eta] /=. rewrite -{2}(idL eta).
  rewrite <-fmap_id. rewrite-> emap1. by destruct (map1 s).
Qed.
Next Obligation.
  move=> C D SP P a b c f g [s eta] /=. rewrite <-compA. rewrite-> fmap_comp.
  rewrite-> emapC. by destruct (mapC f g s).
Qed.

Definition ext_map_full {C D : cat} (SP : Container C D) {P Q : presheaf C}
  (eta : forall c x, pnat (positions SP (c;x)) P -> pnat (positions SP (c;x)) Q)
  (H : forall a b s (f : D b a) (eps : pnat ((positions SP) (b; map f s)) ((positions SP) (a; s))),
      forall omega, eta _ (map f s) (eps ∘ omega) = eps ∘ eta _ s omega)
  :
  pnat (ext SP P) (ext SP Q).
  refine (mk_pnat (fun (d : D) x => _) _). Unshelve. Focus 2.
  cbn. refine (x.1; _). apply eta. exact x.2.
  abstract (move=> a b f [s eps] /=; f_equal; exact: H).
Defined.

Definition ext_map {C D : cat} (SP : Container C D) {P Q : presheaf C} (eta : pnat P Q) :
  pnat (ext SP P) (ext SP Q).
  (*apply: ext_map_full. Unshelve. Focus 2. intros c x eps. exact (eps ∘ eta). move=>/=*. by csimpl. *)
  refine (mk_pnat (fun (d : D) x => _) _). Unshelve. Focus 2.
  cbn. refine (x.1; _). cbn. refine (x.2 ∘ eta).
  abstract (move=> a b f [s eps] /=; f_equal; by csimpl).
Defined.

Lemma ext_map1 {C D : cat} (SP : Container C D) (P : presheaf C) :
  ext_map SP (@cid _ P) = cid.
Proof.
  apply: pnat_eq. fext=> c [s eta] /=. by csimpl.
Qed.

Lemma ext_mapC {C D : cat} (SP : Container C D) {P Q R : presheaf C} (eta : pnat P Q) (eps : pnat Q R) :
  ext_map SP eta ∘ ext_map SP eps = ext_map SP (eta ∘ eps).
Proof.
  apply: pnat_eq. fext=> c [s om] /=. by csimpl.
Qed.

(** ** Initial Algebras of Containers *)

Section InitialAlgebra.
  Variables (C : cat) (SP : Container C C).

  (** ** Overapproximation *)

  Inductive w0 (a : C) :=
  | sup0 (label : shape SP a) (pos : forall b : C, positions SP (a; label) b -> w0 b) : w0 a.

  Definition label0 {a} (w : w0 a) : shape SP a :=
    let (s, _) := w in s.

  Definition pos0 {a} (w : w0 a) : forall b, positions SP (a; label0 w) b -> w0 b :=
    match w as w return forall b, positions SP (a; label0 w) b -> w0 b with sup0 _ F => F end.

  Definition wmap0 {a b} (f : C b a) (w : w0 a) : w0 b :=
    match w with
    | sup0 s F => @sup0 b (map f s) (fun i p => F i (fmap (positions SP) (emap f s) i p))
    end.

  Lemma W_sup_eq (a : C) (s t : shape SP a)
        (F : forall b, positions SP (a; s) b -> w0 b) (G : forall b, positions SP (a; t) b -> w0 b) :
    forall (E : s = t),
      (forall i (p : positions SP (a; s) i),
          F i p = G i (transport E (P := fun x => positions SP (a; x) i) p)) ->
      sup0 F = sup0 G.
  Proof.
    destruct E. cbn. intros H. f_equal. fext=> i p. exact: H.
  Qed.

  Lemma wmap0_id a (w : w0 a) : wmap0 cid w = w.
  Proof.
    case: w => /= s F. apply: W_sup_eq. by csimpl. move=> H i p.
    have->: emap cid s =
            transport (esym H) (P := fun y => el_hom (_; y) (_; s))
                      (@cid (EL (shape SP)) _).
    by apply: sig_eq; destruct H.
    f_equal. destruct H. cbn. by rewrite fmap_id.
  Qed.

  Lemma wmap0_comp (a b c : C) (f : C a b) (g : C b c) (x : w0 c) :
    wmap0 f (wmap0 g x) = wmap0 (f ∘ g) x.
  Proof.
    destruct x as [s F]. move=>/=. apply: W_sup_eq. by csimpl. move=> H i p.
    rewrite-> emapC. have: mapC f g s = H. exact: pi. destruct 1.
    f_equal. destruct (mapC f g s). move=>/=. by rewrite <-fmap_comp.
  Qed.

  Definition W0 : presheaf C :=
    mk_presheaf w0 (@wmap0) wmap0_id wmap0_comp.

  (** ** Cutting it down to the presheaf of hereditarily natural elements *)

  Definition is_natural {a} (w : w0 a) : Prop :=
    forall (b c : C) (f : C c b) (x : positions SP (a; label0 w) b),
      pos0 (map f x) = wmap0 f (pos0 x).

  Fixpoint is_hnatural {a} (w : w0 a) : Prop :=
    is_natural w /\ let (s,F) := w in forall b p, is_hnatural (F b p).

  Lemma wmap_hnatural a b (f : C b a) (w : w0 a) :
    is_hnatural w -> is_hnatural (wmap0 f w).
  Proof.
    destruct w as [s F]. cbn in *. move=> [E wP]. split.
    - move=>/= i j g x. unfold is_natural in E. rewrite <-E. cbn. f_equal. exact: natP.
    - move=> i p. exact: wP.
  Qed.

  Structure wn a := mk_wn { wunfold : w0 a; wisnatural : is_hnatural wunfold }.
  Arguments mk_wn {a} w wP : rename.

  Lemma wn_eq a (x y : wn a) : wunfold x = wunfold y -> x = y.
  Proof. destruct x, y => /= E. destruct E. f_equal. exact: pi. Qed.

  Definition wmap a b (f : C b a) (w : wn a) : wn b :=
    mk_wn (wmap0 f (wunfold w)) (wmap_hnatural f (wisnatural w)).

  Lemma wmap_id a (w : wn a) : wmap cid w = w.
  Proof. apply: wn_eq. exact: wmap0_id. Qed.

  Lemma wmap_comp a b c (f : C a b) (g : C b c) (w : wn c) :
    wmap f (wmap g w) = wmap (f ∘ g) w.
  Proof. apply: wn_eq. exact: wmap0_comp. Qed.

  Definition W : presheaf C :=
    mk_presheaf wn wmap wmap_id wmap_comp.

  (** ** The algebra structure on W *)

  Definition label {a} (w : wn a) : shape SP a :=
    label0 (wunfold w).

  Definition pos {a} (w : wn a) : pnat (positions SP (a; label w)) W.
    refine (mk_pnat (fun b p => mk_wn (@pos0 a (wunfold w) b p) _ : W b) _).
    Unshelve. Focus 2. abstract (by destruct w as [[s F] [H1 H2]]).
    abstract (move=> b c f /= x; apply: wn_eq => /=; by destruct w as [[s F] [H1 H2]]).
  Defined.

  Definition sup {a} (s : shape SP a) (F : pnat (positions SP (a; s)) W) : W a.
    refine (mk_wn (sup0 (fun b p => wunfold (F b p))) _).
    abstract (cbn; split;
      [move=> b c f /= x; by rewrite natP
      |move=> b p; by destruct (F b p)]).
  Defined.

  Definition Sup : pnat (ext SP W) W.
    refine (mk_pnat (fun b (p : ext SP W b) => sup p.2) _).
    abstract (move=> a b f /= [s F] /=; exact: wn_eq).
  Defined.

  (** ** Initiality *)
  Section Initiality.
  Variable (A : presheaf C) (alg : pnat (ext SP A) A).

  (** *** First approximation: the carrier of the graph *)

  Inductive wrec_carrier (a : C) :=
  | wmaps
      (s : shape SP a) (F : forall b, positions SP (a; s) b -> wrec_carrier b)
      (r : A a) : wrec_carrier a.

  Definition wrec_carrier_map (a b : C) (f : C b a) (x : wrec_carrier a) : wrec_carrier b :=
    match x with
    | wmaps s F x =>
      @wmaps b (map f s)
             (fun i p => F i (fmap (positions SP) (emap f s) i p)) (map f x)
    end.

  Lemma wrec_carrier_eq (a : C) (s1 s2 : shape SP a) (r1 r2 : A a)
        (F : forall b, positions SP (a; s1) b -> wrec_carrier b)
        (G : forall b, positions SP (a; s2) b -> wrec_carrier b) :
    forall (E : s1 = s2),
      (forall i (p : positions SP (a; s1) i),
          F i p = G i (transport E (P := fun x => positions SP (a; x) i) p)) ->
      r1 = r2 ->
      wmaps F r1 = wmaps G r2.
  Proof.
    intros E1 H E2. destruct E1, E2. f_equal. fext=> i p. exact: H.
  Qed.

  Lemma wrec_carrier_map_id a (x : wrec_carrier a) :
    wrec_carrier_map cid x = x.
  Proof.
    case: x => s F r. apply: wrec_carrier_eq; try by csimpl.
    move=> H i p. have->: emap cid s =
            transport (esym H) (P := fun y => el_hom (_; y) (_; s))
                      (@cid (EL (shape SP)) _).
    by apply: sig_eq; destruct H.
    f_equal. destruct H. cbn. by rewrite fmap_id.
  Qed.

  Lemma wrec_carrier_map_comp (a b c : C) (f : C a b) (g : C b c) (x : wrec_carrier c) :
    wrec_carrier_map f (wrec_carrier_map g x) = wrec_carrier_map (f ∘ g) x.
  Proof.
    destruct x as [s F r]. move=>/=. apply: wrec_carrier_eq; try by csimpl. move=> H i p.
    rewrite-> emapC. have: mapC f g s = H. exact: pi. destruct 1.
    f_equal. destruct (mapC f g s). move=>/=. by rewrite <-fmap_comp.
  Qed.

  Definition wrec0 : presheaf C :=
    mk_presheaf wrec_carrier wrec_carrier_map wrec_carrier_map_id wrec_carrier_map_comp.

  Fixpoint dom0 (a : C) (g : wrec_carrier a) : w0 a :=
    match g with
    | wmaps s F _ => @sup0 a s (fun b p => dom0 (F b p))
    end.

  Definition pdom0 : pnat wrec0 W0.
    refine (@mk_pnat C wrec0 W0 dom0 _).
    abstract (move=> a b f [s F r]; exact: W_sup_eq).
  Defined.

  Definition codom0 (a : C) (g : wrec_carrier a) : A a :=
    match g with
    | wmaps _ _ x => x
    end.

  Definition pcodom0 : pnat wrec0 A.
    refine (@mk_pnat C wrec0 A codom0 _).
    abstract (by move=> a b f []).
  Defined.

  (** *** Second approximation: Cut it down to the natural part *)

  Definition wrec_label {a} (x : wrec_carrier a) : shape SP a :=
    match x with wmaps s _ _ => s end.

  Definition wrec_pos {a} (x : wrec_carrier a) :
    forall b, positions SP (a; wrec_label x) b -> wrec_carrier b :=
    match x as x return forall b, positions SP (a; wrec_label x) b -> wrec_carrier b with
    | wmaps _ F _ => F
    end.

  Definition wrec_is_natural {a} (x : wrec_carrier a) : Prop :=
    forall (b c : C) (f : C c b) (i : positions SP (a; wrec_label x) b),
      wrec_pos (map f i) = map f (wrec_pos i : wrec0 b).

  Fixpoint wrec_is_hnatural {a} (x : wrec_carrier a) : Prop :=
    wrec_is_natural x /\ let (s,F,r) := x in forall b p, wrec_is_hnatural (F b p).

  Lemma wrec_map_hnatural a b (f : C b a) (x : wrec_carrier a) :
    wrec_is_hnatural x -> wrec_is_hnatural (map f (x : wrec0 a)).
  Proof.
    destruct x as [s F r]. cbn in *. move=> [E wP]. split.
    - move=>/= i j g x. unfold wrec_is_natural in E.
      rewrite <-E. cbn. f_equal. exact: natP.
    - move=> i p. exact: wP.
  Qed.

  Structure wrec1 a := mk_wrec1
    { wrec_unfold : wrec_carrier a; wrecisnatural : wrec_is_hnatural wrec_unfold }.
  Arguments mk_wrec1 {a} w wP : rename.

  Lemma wrec1_eq a (x y : wrec1 a) : wrec_unfold x = wrec_unfold y -> x = y.
  Proof. destruct x, y => /= E. destruct E. f_equal. exact: pi. Qed.

  Definition wrec_map a b (f : C b a) (w : wrec1 a) : wrec1 b :=
    mk_wrec1 (map f (wrec_unfold w : wrec0 a)) (wrec_map_hnatural f (wrecisnatural w)).

  Lemma wrec_map_id a (w : wrec1 a) : wrec_map cid w = w.
  Proof. apply: wrec1_eq => /=. exact: wrec_carrier_map_id. Qed.

  Lemma wrec_map_comp a b c (f : C a b) (g : C b c) (w : wrec1 c) :
    wrec_map f (wrec_map g w) = wrec_map (f ∘ g) w.
  Proof. apply: wrec1_eq => /=. exact: wrec_carrier_map_comp. Qed.

  Definition Wrec1 : presheaf C :=
    mk_presheaf wrec1 wrec_map wrec_map_id wrec_map_comp.

  Lemma dom0_is_hnatural a (x : wrec1 a) :
    is_hnatural (dom0 (wrec_unfold x)).
  Proof.
    case: x. elim=> {a} a s F /= ih r [H1 H2]. split.
    - move=> i j f /= x. unfold wrec_is_natural in H1.
      move: (H1 i j f x) => /=->. by case: (F i x).
    - move=> b p. apply: ih. exact: H2.
  Qed.

  Definition dom1 a (x : wrec1 a) : wn a :=
    mk_wn (dom0 (wrec_unfold x)) (dom0_is_hnatural x).

  Definition pdom1 : pnat Wrec1 W.
    refine (@mk_pnat C Wrec1 W dom1 _).
    abstract (move=> a b f [[s F r] H]; exact: wn_eq).
  Defined.

  Definition codom1 a (x : wrec1 a) : A a :=
    codom0 (wrec_unfold x).

  Definition pcodom1 : pnat Wrec1 A.
    refine (@mk_pnat C Wrec1 A codom1 _).
    abstract (by move=> a b f [[s F r] H]).
  Defined.

  (** *** Third step: Finally, define the graph of the recursor *)

  Program Definition wmaps1 a (s : shape SP a) (F : pnat (positions SP (a; s)) Wrec1) (r : A a) :
    wrec1 a := mk_wrec1 (@wmaps a s (fun b p => wrec_unfold (F b p)) r) _.
  Next Obligation.
    move=> a s F r /=. split; last first. move=> b p. exact: wrecisnatural.
    move=> i j f /= x. by rewrite natP.
  Qed.

  Definition wrec1_label a (x : wrec1 a) : shape SP a :=
    wrec_label (wrec_unfold x).

  Definition wrec1_pos a (x : wrec1 a) :
    pnat (positions SP (a; wrec1_label x)) Wrec1.
  Proof.
    refine (@mk_pnat C (positions SP (a; wrec1_label x)) Wrec1
                     (fun b p => mk_wrec1 (@wrec_pos a (wrec_unfold x) b p) _) _).
    Unshelve. Focus 2.
    abstract (destruct x as [[s F r] [H1 H2]] => /=; exact: H2).
    abstract (move=> b c f i /=; apply: wrec1_eq => /=;
      destruct x as [[s F r] [H1 H2]]; exact: H1).
  Defined.

  Definition wrec_is_graph a (x : wrec1 a) : Prop :=
    alg a (wrec1_label x; wrec1_pos x ∘ pcodom1) = pcodom1 a x.

  Fixpoint wrec_is_hgraph a (x : wrec_carrier a) (H : wrec_is_hnatural x) : Prop :=
    wrec_is_graph (mk_wrec1 x H) /\
    match x as x return wrec_is_hnatural x -> Prop with
    | wmaps _ F _ => fun H => forall i p, @wrec_is_hgraph _ (F i p) (proj2 H i p)
    end H.
  Arguments wrec_is_hgraph {a} x H.

  Definition wrec1_is_hgraph a (x : wrec1 a) : Prop :=
    @wrec_is_hgraph a (wrec_unfold x) (wrecisnatural x).

  Structure wrec_graph a := mk_wrec_graph
    { to_wrec1 : wrec1 a; wrec_graphP : wrec1_is_hgraph to_wrec1 }.

  Program Definition wrec_graph_map a b (f : C b a) (x : wrec_graph a) : wrec_graph b :=
    @mk_wrec_graph b (map f (to_wrec1 x : Wrec1 a)) _.
  Next Obligation.
    move=> a b f [/=-[/=[s F r] H/=]]. unfold wrec1_is_hgraph => /=. unfold wrec_is_graph. cbn.
    move=> [H1 H2]. split; last first.
    - move=> i p. move: (H2 i ((fmap (positions SP) (emap f s)) i p)).
      have E: (proj2 H i ((fmap (positions SP) (emap f s)) i p)) =
              (proj2 (wrec_map_hnatural f H) i p).
      exact: pi. rewrite E. exact.
    - rewrite -{3}H1 /=. rewrite -natP /=. f_equal. f_equal. cbn.
      exact: pnat_eq.
  Qed.

  Definition wrec_graph_dom a : wrec_graph a -> wn a.
    intros H. apply dom1. exact: (to_wrec1 H).
  Defined.

  Lemma wrec_graph_eq a (x y : wrec_graph a) :
    to_wrec1 x = to_wrec1 y -> x = y.
  Proof.
    destruct x, y => /= E. destruct E. f_equal. exact: pi.
  Qed.

  (* TODO: move up? *)
  Definition wn_ind (B : forall a, W a -> Type) :
    (forall a (s : shape SP a) (F : pnat (positions SP (a; s)) W),
        (forall i p, B i (F i p)) -> B a (sup F)) ->
    forall a (x : wn a), B a x.
  Proof.
    move=> H a [w wP]. elim: w wP => {a} a s F ih wP.
    have->:
      {| wunfold := sup0 F; wisnatural := wP |} =
      sup (pos {| wunfold := sup0 F; wisnatural := wP |}) by exact: wn_eq.
    apply: (H a s (@pos a (mk_wn (sup0 F) wP))) => i p. exact: ih.
  Defined.

  Lemma wrec_graph_dom_injective a : injective (@wrec_graph_dom a).
  Proof.
    move=> [[g1 gP1] gP2] g2. unfold wrec_graph_dom => /=.
    unfold dom1 => /= /(ap (@wunfold a))/= E. apply: wrec_graph_eq => /=.
    apply: wrec1_eq => /=. elim: g1 gP1 gP2 g2 E => {a} a s F /= ih r [gP1 gP2].
    unfold wrec1_is_hgraph => /=-[]. unfold wrec_is_graph => /=.
    unfold wrec1_label => /=. unfold codom1 => /= E gP3.
    move=> [[[s' F' r'] /=[gP4 gP5]]]/=. unfold wrec1_is_hgraph.
    move=>/=. unfold wrec_is_graph=>/=. unfold wrec1_label=>/=.
    unfold codom1 => /=-[E' gP6]. move=> dom_eqn.
    have: s = s'. exact: (ap (@label0 a) dom_eqn). move=> EQN. destruct EQN.
    have EQN: forall b p, F b p = F' b p. {
      move=> b p. move: (ih b p). move=>/(_ (gP2 b p)).
      unfold wrec1_is_hgraph => /=. case/(_ _)/Wrap.
      move: (gP3 b p) => /=. have: (proj2 (conj gP1 gP2) b p) = (gP2 b p). exact: pi.
      move->. exact.
      have HH: wrec_is_hgraph (F' b p) (gP5 b p).
      { move: (gP6 b p). have: (proj2 (conj gP4 gP5) b p) = (gP5 b p). exact: pi. move->. exact. }
      move=> /(_ (@mk_wrec_graph b (@mk_wrec1 b (F' b p) (gP5 b p)) HH)).
      cbn. apply. inversion dom_eqn. apply sigT_injective in H0.
      exact: (ap (fun f => f b p) H0). }
    f_equal. fext. exact: EQN.
    rewrite -E -E' /=. f_equal. f_equal. apply: pnat_eq.
    fext=> b p /=. unfold codom1 => /=. f_equal. exact: EQN.
  Qed.

  Structure wrec_image a (x : wn a) :=
    mk_wrec_image
      { wrec_image_graph :> wrec_graph a
      ; wrec_image_dom : wrec_graph_dom wrec_image_graph = x
      (*; wrec_image_uniq : forall z : wrec_graph a,
          wrec_graph_dom z = x -> wrec_image_graph = z*)
      }.

  Definition codom2 a (x : wrec_graph a) : A a :=
    codom1 (to_wrec1 x).

  Program Definition wrec_image_codom a (s : shape SP a) (F : pnat (positions SP (a; s)) W)
    (img : forall (i : C) (p : (positions SP (a; s)) i), wrec_image (F i p)) :
    pnat (positions SP (a; s)) A :=
    mk_pnat (fun b p => codom2 (img b p)) _.
  Next Obligation.
    move=> a s F img b c f x /=.
    case: (img b x) => g1 eqn1 /=.
    case: (img c (map f x)) => g2 eqn2 /=.
    rewrite -(natP pcodom1) /=. unfold codom2. f_equal.
    rewrite (@wrec_graph_dom_injective c g2 (wrec_graph_map f g1)) //.
    unfold wrec_graph_map => /=.
    unfold wrec_graph_dom in * => /=.
    move: (natP pdom1 f (to_wrec1 g1)) => /=->.
    by rewrite eqn1 eqn2 natP.
  Qed.

  Program Definition wrec_image_wrec a (s : shape SP a) (F : pnat (positions SP (a; s)) W)
    (img : forall (i : C) (p : (positions SP (a; s)) i), wrec_image (F i p)) :
    pnat (positions SP (a; s)) Wrec1 :=
    mk_pnat (fun b p => to_wrec1 (img b p)) _.
  Next Obligation.
    move=> a s F img b c f x /=.
    case: (img b x) => g1 eqn1 /=.
    case: (img c (map f x)) => g2 eqn2 /=.
    rewrite (@wrec_graph_dom_injective c g2 (wrec_graph_map f g1)) //.
    unfold wrec_graph_map => /=.
    unfold wrec_graph_dom in * => /=.
    move: (natP pdom1 f (to_wrec1 g1)) => /=->.
    by rewrite eqn1 eqn2 natP.
  Qed.

  Definition wrec_image_rec a (s : shape SP a) (F : pnat (positions SP (a; s)) W)
    (img : forall (i : C) (p : (positions SP (a; s)) i), wrec_image (F i p)) :
    wrec1 a :=
    @wmaps1 a s
            (@wrec_image_wrec a s F img)
            (alg a (s; @wrec_image_codom a s F img)).

  Lemma wrec_image_rec_graph a (s : shape SP a) (F : pnat (positions SP (a; s)) W)
    (img : forall (i : C) (p : (positions SP (a; s)) i), wrec_image (F i p)) :
    wrec1_is_hgraph (@wrec_image_rec a s F img).
  Proof.
    cbn. split.
    - unfold wrec_is_graph => /=. unfold wrec1_label => /=.
      unfold codom1 => /=. f_equal. f_equal. apply: pnat_eq.
      fext=> i p /=. unfold codom2 => /=. f_equal.
        by apply: wrec1_eq => /=.
    - move=> i p. move: (wrec_graphP (img i p)).
      unfold wrec1_is_hgraph.
      have<-: (wrecisnatural (to_wrec1 (img i p))) =
              (proj2 (wmaps1_obligation_1 (wrec_image_wrec img) (alg a (s; wrec_image_codom img))) i p).
      exact: pi. exact.
  Qed.

  Lemma wrec_graph_dom_surjective a (x : wn a) : wrec_image x.
  Proof.
    move: a x. apply: wn_ind => a s F ih.
    exists (@mk_wrec_graph a (@wrec_image_rec a s F ih) (@wrec_image_rec_graph a s F ih)).
    apply: wn_eq => /=. f_equal. fext=> b p.
    case: (ih b p) => g1 eqn /=. by rewrite -eqn.
  Qed.

  Lemma wrec_graph_dom_is_equiv a :
    is_equiv (@wrec_graph_dom a).
  Proof.
    refine (mk_is_equiv (fun x : wn a => wrec_image_graph (wrec_graph_dom_surjective x)) _ _).
    - move=> x. apply: wrec_graph_dom_injective => /=.
      destruct (wrec_graph_dom_surjective (wrec_graph_dom x)) => /=.
        by rewrite wrec_image_dom0.
    - move=> x. destruct (wrec_graph_dom_surjective x) => /=.
        by rewrite wrec_image_dom0.
  Qed.

  Definition wrec_graph_proj a :
    wrec_graph a <~> wn a :=
    mk_equiv (@wrec_graph_dom a) (@wrec_graph_dom_is_equiv a).

  Definition to_wrec_graph a (x : wn a) : wrec_graph a :=
    (@wrec_graph_proj a)^-1 x.

  (* Now:
     - wrec_graph is a presheaf
     - wrec_graph_dom, wrec_graph_codom are natural transformations
     - since wrec_graph_dom is componentwise an equivalence it is a natural isomorphism
     - we define the natural transformation

         wrec : pnat W A
         wrec := wrec_graph_dom^-1 ∘ wrec_graph_codom

     - Furthermore this is an ext SP-algebra homomorphism:
         ext_map (ext SP) wrec ∘ alg = Sup ∘ wrec

         ext_map (ext SP) (wrec_graph_dom^-1 ∘ wrec_graph_codom) ∘ alg
       = E (wrec_graph_dom^-1) ∘ E wrec_graph_codom ∘ alg
       = (E wrec_graph_dom)^-1 ∘ E wrec_graph_codom

      STS
        E wrec_graph_codom ∘ alg = E wrec_graph_dom ∘ Sup ∘ wrec

      STS
        E wrec_graph_dom ∘ Sup ∘ wrec_graph_dom^-1 = id

      STS
        E wrec_graph_codom ∘ alg = wrec

     - Finally, we also need to show uniqueness.
   *)

  Lemma wrec_graph_map_id a (x : wrec_graph a) :
    wrec_graph_map cid x = x.
  Proof.
    apply: wrec_graph_eq => /=. exact: wrec_map_id.
  Qed.

  Lemma wrec_graph_map_comp a b c (f : C a b) (g : C b c) (x : wrec_graph c) :
    wrec_graph_map f (wrec_graph_map g x) = wrec_graph_map (f ∘ g) x.
  Proof.
    apply: wrec_graph_eq => /=. exact: wrec_map_comp.
  Qed.

  Definition Wrec := mk_presheaf wrec_graph wrec_graph_map wrec_graph_map_id wrec_graph_map_comp.

  Program Definition Wrec_dom : pnat Wrec W :=
    mk_pnat (fun b x => wrec_graph_dom x) _.
  Next Obligation.
    move=> a b f x /=.
    move: (@natP _ _ _ pdom1 a b f (to_wrec1 x)) => /=.
    unfold wrec_graph_dom. by move<-.
  Qed.

  Program Definition Wrec_codom : pnat Wrec A :=
    mk_pnat (fun b x => codom2 x) _.
  Next Obligation.
    move=> a b f x /=. unfold codom2. unfold codom1.
    move=> /=. destruct (to_wrec1 x) as [g H] => /=.
      by destruct g.
  Qed.

  Program Definition to_Wrec : pnat W Wrec :=
    mk_pnat (fun b x => (@wrec_graph_proj b)^-1 x) _.
  Next Obligation.
    move=> a b f x /=. apply: wrec_graph_dom_injective.
    move: (equiv_retr (@wrec_graph_proj b) (wmap f x)).
    move=> /=. move=> HH. transitivity (wmap f x). exact: HH. move=>{HH}.
    move: (@natP _ _ _ Wrec_dom a b f ((@wrec_graph_proj a)^-1 x)) => /= HH.
    rewrite->HH. f_equal. symmetry. exact (equiv_retr (@wrec_graph_proj a) x).
  Qed.

  Definition wrec : pnat W A :=
    to_Wrec ∘ Wrec_codom.

  Lemma wrec_beta :
    ext_map SP wrec ∘ alg = Sup ∘ wrec.
  Proof.
    apply: pnat_eq. fext=> a x /=.
    move eqn:((@wrec_graph_proj a)^-1 (sup x.2)) => g.
    destruct g as [g gP] => /=. unfold codom2 => /=.
    move: (ap (@wrec_graph_proj a) eqn). rewrite equiv_retr => /={eqn} eqn.
    destruct g as [[s F r] H]. cbn in gP. destruct gP as [gP1 gP2].
    unfold wrec_is_graph in gP1. move: (gP1) => /=.
    unfold codom1 => /=<-/=. f_equal.
    unfold wrec1_label => /=. unfold wrec_graph_dom in eqn.
    unfold dom1 in eqn. cbn in eqn.
    move: (ap (@wunfold _) eqn) => /={eqn} eqn.
    move: (ap (@label0 _) eqn) => /= E. destruct E. f_equal.
    apply: pnat_eq. fext=> b p /=.
    unfold codom2 => /=. f_equal. apply: wrec1_eq => /=.
  Admitted.

  End Initiality.

  (* Uniqueness *)
  Section Uniqueness.
  Variable (A : presheaf C).
  Variable (alg : pnat (ext SP A) A).
  Variable (η : pnat W A).
  Variable (β : ext_map SP η ∘ alg = Sup ∘ η).

  (* Idea: build a wrec_graph for η *)

  Definition eta_wrec_carrier : pnat W (@Wrec A alg).
  apply wrec.
  Definition eta_wrec_carrier a (x : wn a) : wrec_carrier a.
    move: a x. apply wn_ind. move=> a s p ih.
    exists s. exact ih. apply alg. exists s. refine (p ∘ _). exact η.
  Defined.

  Lemma eta_wrec_carrier_is_natural a (x : wn a) :
    wrec_is_hnatural (eta_wrec_carrier x).
  Proof.
    move: a x. apply: wn_ind => a s p ih.
    move=> /=. unfold ssr_have. cbn.

    split. unfold wrec_is_hnatural.


    :=
    match w0 with
    |
  Lemma wrec_eta (η : pnat W A) :
    (ext_map SP η ∘ alg = Sup ∘ η) -> η = wrec.
  Proof.

    unfold codom1 => /=.
    dependent inversion eqn. move: (ap (@pos0 a) eqn).



      Set Printing All. f_equal. destruct (ih b p).
      
      
    (* Construct the result *)
    have: A a.
    apply: alg. hnf. exists s.
    apply: mk_pnat. Unshelve. Focus 3.
    intros b p. destruct (ih b p). destruct x as [x xP]. apply (codom1 x).
    Focus 2. cbn. intros b c f x. destruct (ih b x). destruct (ih c (map f x)).
    destruct x1 as [g1 gP1]. destruct x0 as [g2 gP2]. rewrite -(natP pcodom1) /=.
    f_equal. destruct a1 as [H HI]. move: (HI (wrec_graph_map f (mk_wrec_graph gP2))).
    unfold wrec_graph_dom in * => /=. apply. destruct a0 as [HH HHH].
    cbn in HH. cbn in H. move: (natP pdom1 f g2). cbn. move->.
    rewrite HH. by rewrite natP.


  Definition wrec' a (x : w0 a) :
    is_hnatural x -> 

  Lemma wrec_graph_dom_injective a : injective (@wrec_graph_dom a).
  Proof.
    case=> -[/=]. elim=> {a} /= a s F ih r H p [[[s' F' r'] H'] p'] E.
    apply sig_eq => /=. apply: wrec1_eq => /=. unfold wrec_graph_dom in E. cbn in E.
    unfold dom1 in E. cbn in E. apply sigT_eq_pr1 in E.

  Definition wrec' a (x : wn a) : { x : wrec1 a | wrec_graph x }.
  Proof.
    destruct x as [x H]. elim: x H => {a} a s F ih /=[H1 H2].
    eexists. Unshelve. Focus 2. apply: (mk_wrec1 (@wmaps a s _ _) _).
    Unshelve. Focus 2. intros b p. exact (wrec_unfold (ih b p (H2 b p)).1).
    refine (alg a (s; _ ∘ pcodom1)).
    apply: (mk_wrec1
    apply: alg. exists 
    refine ((wrec1_pos _ ∘ pcodom1) a _).
    (exists s.

  End Initiality.

End InitialAlgebra.

(** ** Presheaf W-types *)

Section InitialAlgebra.
  Variables (C : cat) (SP : Container C C).

  Polymorphic Cumulative Inductive W (c : C) : Type :=
  | sup (s : shape SP c) (p : forall i : C, positions SP (c; s) i -> W i) : W c.

  Definition W_map (a b : C) (f : C b a) (x : W a) : W b.
    destruct x as [s F]. eapply sup. Unshelve. Focus 2.
    exact (map f s). intros i p. apply F. apply (fmap (positions SP) (emap f s)). exact p.
  Defined.

  Definition W_nat_at (a : C) (x : W a) : Prop :=
    match x with
    | sup s F => forall (i j : C) (f : C j i) (x : positions SP (a; s) i),
        F j (map f x) = W_map f (F i x)
    end.

  Fixpoint W_nat (a : C) (x : W a) : Prop :=
    W_nat_at x /\ match x with
                  | sup s F => forall i p, W_nat (F i p)
                  end.

  Lemma W_nat_map (a b : C) (f : C b a) (x : W a) :
    W_nat x -> W_nat (W_map f x).
  Proof.
    destruct x as [s F]. cbn in *. move=> [E wP]. split.
    - move=> i j g x. rewrite <-E. f_equal. exact: natP.
    - move=> i p. exact: wP.
  Qed.

  Lemma W_sup_eq (a : C) (s t : shape SP a) (F : forall i, positions SP (a; s) i -> W i) (G : forall i, positions SP (a; t) i -> W i) :
    forall (E : s = t),
      (forall i (p : positions SP (a; s) i),
          F i p = G i (transport E (P := fun x => positions SP (a; x) i) p)) ->
      sup F = sup G.
  Proof.
    destruct E. cbn. intros H. f_equal. fext=> i p. exact: H.
  Qed.

  (* same proofs as above? *)
  Lemma W_map_id (a : C) (x : W a) :
    W_map cid x = x.
  Proof.
    destruct x as [s F]. move=>/=. apply: W_sup_eq. by csimpl.
    move=> H i p. have HH: emap cid s = transport (esym H) (P := fun y => el_hom (_; y) (_; s)) (@cid (EL (shape SP)) _). apply: sig_eq. cbn. by destruct H.
    rewrite HH. f_equal. clear HH. destruct H. cbn. by rewrite fmap_id.
  Qed.

  Lemma W_map_comp (a b c : C) (f : C a b) (g : C b c) (x : W c) :
    W_map f (W_map g x) = W_map (f ∘ g) x.
  Proof.
    destruct x as [s F]. move=>/=. apply: W_sup_eq. by csimpl. move=> H i p.
    rewrite-> emapC. have: mapC f g s = H. exact: pi. destruct 1.
    f_equal. destruct (mapC f g s). move=>/=. by rewrite <-fmap_comp.
  Qed.

  Program Definition WP : presheaf C :=
    mk_presheaf
      (fun a => { w : W a | W_nat w })
      (fun a b f w => exist (@W_nat b) (W_map f w.1) (W_nat_map f w.2)) _ _.
  Next Obligation.
    move=> a [w wP] /=. apply sig_eq => /=. exact: W_map_id.
  Qed.
  Next Obligation.
    move=> a b c f g /= [w wP] /=. apply: sig_eq => /=. exact: W_map_comp.
  Qed.

  Program Definition wfold : pnat (ext SP WP) WP.
  Proof.
    refine (mk_pnat _ _). Unshelve. Focus 2. cbn.
    refine (fun a wp => exist _ (@sup a wp.1 (fun i p => (wp.2 i p).1)) _).
    abstract (cbn; split; [intros i j f x; by rewrite natP|intros i p; exact (svalP (wp.2 i p))]).
    abstract (move=> a b f /=[s eta]; exact: sig_eq).
  Defined.

  Program Definition wunfold : pnat WP (ext SP WP).
  Proof.
    refine (mk_pnat _ _). Unshelve. Focus 2. cbn.
    intros a [[s F] wP]. exists s.
    refine (mk_pnat _ _). Unshelve. Focus 2. cbn.
    intros b p. exists (F b p). apply wP.2.
    cbn. intros c d f x. apply sig_eq => /=. apply wP.1.
    intros a b f [[s F] wP]. move=>/=. f_equal.
    apply: pnat_eq. cbn. fext=> c p. by apply sig_eq.
  Defined.

  Lemma WP_ind (P : forall c, WP c -> Type) :
    (forall c (s : shape SP c) (F : pnat (positions SP (c; s)) WP),
        (forall i p, P i (F i p)) -> P c (wfold c (s; F))) ->
    (forall c x, P c x).
  Proof.
    intros H c [w wP]. revert wP. induction w as [c s F ih]. intros [E wP].
    move: (H c s). move: (mk_pnat (fun i p => (exist (@W_nat _) (F i p) (wP i p)) : WP i)).
  Admitted.
(*
    move=> /(_ (fun (a b : C) (f : C b a) (x0 : ((positions SP) (c; s)) a) => sig_eq (E a b f x0))).

    cbn. case/(_ _)/Wrap. intros a b f x. apply sig_eq. cbn. exact (E a b f x). Show Proof.
        
    ) -> True.*)

  Section Recursion.
    Variable D : presheaf C.
    Variable alg : pnat (ext SP D) D.

    Definition sh (a : C) (w : W a) : shape SP a :=
      match w with sup s _ => s end.

    Definition wrec_below (a : C) (w : W a) :
      { eta : pnat (positions SP (a; sh w)) P -> pnat (positions SP (a; sh w)) Q)
      & forall a b s (f : D b a) (eps : pnat ((positions SP) (b; map f s)) ((positions SP) (a; s))),
      forall omega, eta _ (map f s) (eps ∘ omega) = eps ∘ eta _ s omega
    Definition wrec_below (a : C) (w : W a) :
      pnat (positions SP (a; sh w)) D.
    Proof.
      induction w as [a s F ih]. refine (_ ∘ alg). apply: mk_pnat.
      Unshelve. Focus 2. intros b. cbn. intros i. exists (sh (F b i)).
      


      apply: ih. cbn. intros b c f x.
      exists s.
    Qed.

    Definition elem (c d : C) (w : W c) (w' : W d) : Type :=
      match w with
      | sup s F => { i : positions SP (c; s) d | F d i = w' }
      end.

    Lemma elem_map c d1 d2 (f : C d2 d1) (w : W c) (w' : W d1) :
      W_nat_at w -> elem w w' -> elem w (W_map f w').
    Proof.
      case: w => s F /=H [i P]. exists (map f i). abstract (subst; exact: H).
    Defined.

    Program Definition Elem {c} (w : WP c) : presheaf C :=
      mk_presheaf (fun d => { w' : _ & @elem c d w.1 w' }) _ _ _.
    Next Obligation.
      intros c w a b f. cbn. intros p. exists (W_map f p.1). apply: elem_map; last first. exact: p.2.
      destruct w. cbn. destruct x. exact w.1.
    Defined.
    Next Obligation.
    Admitted.
    Next Obligation.
    Admitted.

    Definition wrec_elem (c : C) (w : WP c) :
      pnat (Elem w) D.
    Proof.
      destruct w as [w wP]. revert wP. induction w as [c s F ih]. intros wP.
      refine (_ ∘ alg).

    Definition wrec_elem (c : C) (w : W c) :
      W_nat w -> { eta : forall d (w' : W d), elem w w' -> D d &
        forall d1 d2 (f : C d2 d1) (w' : W d1) (H1 : elem w w') (H2 : elem w (W_map f w')),
          map f (eta _ _ H1) = eta _ _ H2 }.
    Proof.
      induction w as [c s F ih]. intros [w_nat_at w_nat]. eexists.
      Unshelve. Focus 2. intros d w' H. apply alg. destruct w' as [s' F'].
      exists s'. destruct H as [p eqn]. destruct (ih _ p (w_nat _ p)).
      apply: mk_pnat. Unshelve. Focus 2. intros a q. eapply x. rewrite eqn. eexists. reflexivity.
      Unshelve. Focus 2. exact q.
      cbn. intros a b f y. erewrite e. unfold eq_rect_r. unfold eq_rect. admit.

      intros. cbn. destruct w' as [t G]. cbn. rewrite<- natP. f_equal.
      cbn. f_equal. apply: pnat_eq. fext=> i q /=. destruct H1.
      destruct H2 as [H20 H2]. assert (H20 = map f x). admit. subst.
      move: (ih _ x (w_nat _ x)).
      d


      Unshelve. Focus 2. apply elem_map. apply (w_nat d p).
      intros d w' [p eqn]. destruct (ih _ p (w_nat _ p)).
      eapply x. eexists.


    Definition sh {c : C} (w : WP c) : shape SP c. destruct w as [[s _] _]. exact s. Defined.

    Definition pos {c : C} (w : WP c) : pnat (positions SP (c; sh w)) WP.
      destruct w as [[s F] [E wP]]. apply: mk_pnat. Unshelve. Focus 2.
      intros a. move=>/=. intros p. exists (F a p). apply (wP a p).
      intros a b f. move=>/=. intros x. apply sig_eq. cbn. exact: E.
    Defined.

    Definition wsh (c : C) (w : W c) : shape SP c. destruct w as [s _]. exact s. Defined.


    Definition wrec' (c : C) (w : WP c) :
      D c * (pnat (positions SP (c; sh w)) D).
    Proof.
      destruct w as [w wP]. revert c w wP. fix ih 2. intros c [s F] wP.
      split. apply alg. exists s. destruct wP as [_ wP].
    Definition wrec' (c : C) (w : WP c) :
      pnat (positions SP (c; sh w)) D.
    Proof.
      destruct w as [w wP]. revert c w wP. fix ih 2. intros c [s F] wP.
      refine (_ ∘ alg). refine (pos _ ∘ _). refine (wunfold ∘ _). apply ext_map.
      refine (_ ∘ alg). refine (wunfold ∘ _). exists

    Definition wrec' (c : C) (w : W c) :
      W_nat w -> pnat (positions SP (c; wsh w)) D.
      induction w as [c s F ih]. cbn=> -[H1 H2].
      refine (_ ∘ alg).
      apply: mk_pnat. Unshelve. Focus 2.
      intros a p. exists (wsh (F a p)). exact: ih.
      intros a b f x. move=>/=. Check ih b (map f x) (H2 b (map f x)). cbn. apply: sigT_eq. rewrite H1. by destruct (F a x).
      intros H. apply: pnat_eq. fext=> c' /= y. destruct (ih a x (H2 a x)). cbn.
    Admitted.


    Definition wrec' (c : C) (w : WP c) :
      pnat (positions SP (c; sh w)) D.
    Proof.
      destruct w as [w wP]. revert wP. induction w as [c s F ih]. intros wP.
      refine (_ ∘ alg). apply: mk_pnat. Unshelve. Focus 2.
      intros a. cbn. intros p. exists (sh (exist _ (F a p) (wP.2 a p))).
      apply ih. intros a b f x. destruct wP as [E wP]. cbn in E.
      move=>/=.
      apply: sigT_eq. cbn. destruct wP.
      change (sh (exist (W_nat (a:=b)) (F b (map f x)) (wP.2 b (map f x)));
              ih b (map f x) (wP.2 b (map f x))) =
                 map f (sh (exist (W_nat (a:=a)) (F a x) (wP.2 a x)); ih a x (wP.2 a x))).

      move=>/=. destruct (F a x). move: (sh (exist (W_nat (a:=a0)) (F a0 p) (wP.2 a0 p))). intros x.
      destruct (F a p) as [t G]. exists t.
      apply ih.
      exists (shape (F a p)).
      refine (pos _ ∘ _).

      { eta : forall d (p : positions SP (c; sh w) d), pos w d p -> D d & True }. 
    .

    (* Can't do induction on W, need to have a type of ordinal approximations to W

      Ord : Type := Sup (A : Type) : (A -> Ord) : Ord.

      Fixpoint W_approx (o : Ord) (c : C) : Type :=
       { s : shape SP c & forall (p : o.type), pnat (positions SP (c; s)) (W_approx (o.down p)) }.

      where pnat has to be expanded into a separate definition again...

      We have a function forall w : W c, W_approx (subtreees w) c

      Then we can construct the recursor forall o, W_approx o -> D by induction on o.
      Show that it is an SP algebra homomorphism and unique.
     *)

    Polymorphic Cumulative Inductive Ord : Type :=
    | Limit (A : Type) : (A -> Ord) -> Ord.

    Fixpoint W_approx (o : Ord) (c : C) : Type :=
      match o with
      | Limit A down => { s : shape SP c & forall (c' : C), positions SP (c; s) c' -> { i : A & W_approx (down i) c' } }
      end.

    Fixpoint subtrees (c : C) (w : W c) : Ord :=
      match w with
      | sup s F => @Limit { i : C & positions SP (c; s) i } (fun p => subtrees (F p.1 p.2))
      end.

    Definition W_to_approx (c : C) (w : W c) : W_approx (subtrees w) c.
      induction w. cbn. exists s. intros c' q. eexists. Unshelve.
      Focus 2. exists c'. exact q. cbn. exact: X.
    Defined.

    Definition W_approx_map (o : Ord) (c d : C) (f : C d c) (x : W_approx o c) : W_approx o d.
      destruct o as [A down]. cbn in *. destruct x as [s F]. exists (map f s).
      intros c' p. apply (F c' (fmap (positions SP) (emap f s) _ p)).
    Defined.

    Definition wrec_approx (o : Ord) :
      { eta : forall c, W_approx o c -> D c
      & forall c d (f : C d c) (x : W_approx o c),
          eta d (W_approx_map f x) = map f (eta c x)
      }.
    Proof.
      induction o. cbn in *. eexists. Unshelve. Focus 2. intros c [s F].
      apply alg. cbn. exists s. apply: mk_pnat. Unshelve. Focus 2.
      intros c' p. destruct (F c' p) as [i w]. destruct (X i). apply x. exact w.
      intros a b f x. cbn. destruct (F a x) as [i w]. destruct (X i) as [eta etaP].
      rewrite -etaP. unfold W_approx_map. destruct (o i). cbn in *.
      intros c. exists

    Inductive wrec_graph (a : C) (w : WP a) : D a -> Prop :=
    | Wrec_graph (eta : pnat WP D) :
        (forall b w', @wrec_graph b w' (eta b w')) ->
        @wrec_graph a w (alg _ (ext_map SP eta _ (wunfold a w))).

    Lemma wrec_graph_functional (a : C) (w : WP a) (x y : D a) :
      wrec_graph w x -> wrec_graph w y -> x = y.
    Proof.
      move=> H. elim: H y => {a w x} a w eta H1 ih y [eps H2].
      f_equal. f_equal. f_equal. apply pnat_eq. fext=> b w'.
      apply: ih. exact: H2.
    Qed.

        (forall b 
        wunfold w alg d
        w
    .

    Definition wrec : pnat WP D.
      refine (mk_pnat _ _). Unshelve. Focus 2.
      apply: WP_ind. intros c s eta ih. apply alg. exists s.

      cbn.
      intros a [w wP]. induction w. apply: alg. cbn. exists s.
      refine (mk_pnat _ _). Unshelve. Focus 2. intros a q. apply (X a q). apply wP.2.
      cbn. intros a b f x. Unshelve.
      exact: p.
End InitialAlgebra.