Require Import base ord.

Polymorphic Class monad_ret (m : Type -> Type) :=
  ret : forall {X}, X -> m X.
Polymorphic Class monad_bind (m : Type -> Type) :=
  bind : forall {X Y}, m X -> (X -> m Y) -> m Y.
Polymorphic Class monad (m : Type -> Type) `{monad_ret m} `{monad_bind m} :=
  { bind1x : forall X Y (x : X) (f : X -> m Y),
      bind (ret x) f = f x
  ; bindx1 : forall X (x : m X),
      bind x (@ret m _ X) = x
  ; bindA : forall X Y Z (x : m X) (f : X -> m Y) (g : Y -> m Z),
      bind (bind x f) g = bind x (fun a => bind (f a) g)
  }.

Notation "x >>= f" := (@bind _ _ _ _ x f) (at level 55, right associativity) : monad_scope.
Notation "f >=> g" := (fun x => bind (f x) g) (at level 55, right associativity) : monad_scope.
Open Scope monad_scope.

Section Powerset.
  Definition B (X : Type) := X -> Prop.

  Global Instance: monad_ret B := @eq.
  Global Instance: monad_bind B := fun X Y P F y => exists2 x : X, P x & F x y.

  Global Instance: @monad B _ _.
  Proof.
    constructor.
    - move=> X Y x f. fext=> y. apply: pext. by case=> x' <-. move=> p. by exists x.
    - move=> X P. fext=> y. apply: pext. by case=> x px <-. move=> py. by exists y.
    - move=> X Y Z P F G. fext=> z. apply: pext; cbv; by firstorder.
  Qed.

  Global Instance: forall X Y, Proper (ord_op ++> ord_op ++> ord_op) (@bind B _ X Y).
  Proof.
    move=> X Y P1 P2 lep F1 F2 lef y /=. case=> x /lep px /lef fxy. by exists x.
  Qed.
End Powerset.

Section Dijkstra.
  Definition D (X : Type) : Type := {mono (X -> Prop) -> Prop}.

  Global Instance: monad_ret D := fun X x => mk_mfun (fun (Q : X -> Prop) => Q x) _.
  Proof. move=> /=P Q. exact. Defined.

  Global Instance: monad_bind D := fun X Y F G => mk_mfun (fun (Q : Y -> Prop) => F (fun x => G x Q)) _.
  Proof. move=> /=P Q le_P_Q. apply: mono => x. exact: mono. Defined.

  Global Instance: @monad D _ _.
  Proof.
    constructor.
    - move=> X Y x f. exact: mfun_eq.
    - move=> X f. exact: mfun_eq.
    - move=> X Y Z x f g. exact: mfun_eq.
  Qed.

  Global Instance: forall X Y, Proper (ord_op ++> ord_op ++> ord_op) (@bind D _ X Y).
  Proof.
    move=> X Y P1 P2 lep F1 F2 lef Q. unfold bind=>/=. rewrite (lep _). apply: mono.
    move=> x. exact: lef.
  Qed.

  Lemma dijkstra_bind_alternative X Y (W : D X) (F : X -> D Y) (Q : Y -> Prop) :
    (W >>= F) Q = exists2 P : X -> Prop, W P & forall x, P x -> F x Q.
  Proof.
    apply: pext. move=> h. by exists (fun x => F x Q). case=> P wP pF.
    move: wP. apply: (monotone_monofun W). exact: pF.
  Qed.
End Dijkstra.

Require Import tower.tarski.

Definition lfp {X : clat} (f : X -> X) : X :=
  lfp_above f ⊥.

Section Lifting.
  Program Definition lift {X} (P : B X) : D X :=
    mk_mfun (fun Q => P ≤ Q :> (X -> Prop)) _.
  Next Obligation. cbv in *. firstorder. Qed.

  Lemma lift_ret X (x : X) :
    lift (ret x) = ret x.
  Proof.
    apply: mfun_eq. fext=> Q /=. apply: pext. exact. move=> qx y <-. exact: qx.
  Qed.

  Lemma lift_bind X Y (x : B X) (f : X -> B Y) :
    lift (x >>= f) = lift x >>= (lift \o f).
  Proof.
    apply: mfun_eq; fext=> Q /=. apply: pext.
    - move=> h a xa b fab. apply: h. by exists a.
    - move=> h b [a xa fab]. exact: h fab.
  Qed.

  Global Instance: forall X, Proper (ord_op --> ord_op) (@lift X).
  Proof.
    move=> X /= P Q le_P_Q R /= h x qx. apply: h. cbv in le_P_Q. exact: le_P_Q.
  Qed.

  Definition unlift {X} (W : D X) : B X :=
    fun x => forall Q, W Q -> Q x.

  Lemma lift_unlift_adj {X} (P : B X) (W : D X) :
    (W ≤ lift P) <-> (P ≤ unlift W).
  Proof.
    split.
    - move=> h x /= px Q wq. exact: h.
    - move=> h Q wq /=. rewrite h => x. exact.
  Qed.

  Lemma lift_sup X I (F : I -> B X) :
    lift (sup F) = inf (lift \o F).
  Proof.
    apply: le_eq.
    - focus=> i /=. apply/lift_unlift_adj => x fi. rewrite/lift/unlift/= => Q.
      apply. rewrite prod_exE prop_exE. by exists i.
    - apply/lift_unlift_adj. focus=> i. hnf=> x. hnf=> fi. rewrite/lift/unlift/= => Q.
      apply. exact fi.
  Qed.

  Lemma unlift_ret X (x : X) :
    unlift (ret x) = ret x.
  Proof.
    fext=> y. apply: pext. exact. by move=>-> Q/=.
  Qed.

  Lemma unlift_lift_retr X (P : B X) :
    unlift (lift P) = P.
  Proof.
    fext=> x. apply: pext. exact. move=> px Q. exact.
  Qed.

  (*Variables (X Y : Type) (F : (X -> B Y) -> (X -> B Y)) (G : (X -> D Y) -> X -> D Y).
  Hypothesis (F_mono : monotone F) (G_mono : monotone G).

  Hypothesis FG_morphism :
    forall (f : X -> B Y), lift \o (F f) = G (lift \o f).

  Lemma F_def (f : X -> B Y) :
    F f = unlift \o G (lift \o f).
  Proof.
    rewrite -FG_morphism. fext=> x y /=. by rewrite unlift_lift_retr.
  Qed.

  Lemma F_fp (g : fp G) :
    lfp F ≤ unlift \o point g.
  Proof.
    apply: lfp_above_ind => //. rewrite F_def -{2}fpP. hnf=> x /=.
    have: (Proper (ord_op --> ord_op) unlift). admit. apply.
    unfold flip. apply: G_mono. hnf=> {x} x. hnf=>/= Q. hnf=> h.
    hnf=> y. hnf. rewrite/unlift. apply. exact: h.
  Admitted.

  Lemma F_fp' (g : fp G) x y Q :
    point g x Q -> lfp F x y -> Q y.
  Proof.
    move=> gxQ /(@F_fp g). apply. exact: gxQ.
  Qed.

  Lemma deterministic_lift_unlift (W : D X) :
    (exists x, forall Q, W Q = Q x) ->
    lift (unlift W) = W.
  Proof.
    move=> H. apply: le_eq. rewrite/lift/unlift. hnf=>/=Q. hnf=> h.
    case: H => x0 eqn. rewrite eqn. apply: h. move=> Q0. rewrite eqn. trivial.

    move=> Q. hnf=> h. hnf=> x. apply. exact: h.
  Qed.*)
End Lifting.
